@@ -174,6 +174,7 @@ def add_standard_builtins(engine, b=None, s=None, sp=None):
     engine.add_builtin('use_module', 2, b(_builtin_use_module2))
     engine.add_builtin('module', 2, b(_builtin_module))
 
+    engine.add_builtin('once', 1, _builtin_call)
     engine.add_builtin('call', 1, _builtin_call)
     engine.add_builtin('call_nc', 1, _builtin_call_nc)
     engine.add_builtin('try_call', 1, _builtin_try_call)
@@ -1275,13 +1276,6 @@ def _atom_to_filename(atom):
     return atomstr
 
 
-class ConsultError(GroundingError):
-    """Error during consult"""
-
-    def __init__(self, message, location):
-        GroundingError.__init__(self, message, location)
-
-
 def _builtin_consult_as_list(op1, op2, **kwdargs):
     """Implementation of consult/1 using list notation.
 
@@ -1308,33 +1302,8 @@ def _builtin_consult(filename, database=None, engine=None, **kwdargs):
    :param kwdargs: additional arguments
    :return: True
     """
-    root = database.source_root
-    if filename.location:
-        source_root = database.source_files[filename.location[0]]
-        if source_root:
-            root = os.path.dirname(source_root)
     check_mode((filename,), ['a'], functor='consult', **kwdargs)
-    filename = _atom_to_filename(filename)
-    if not os.path.exists(filename):
-        filename = os.path.join(root, filename)
-        if not os.path.exists(filename):
-            filename += '.pl'
-        if not os.path.exists(filename):
-            raise ConsultError(message="Consult: file not found '%s'" % filename,
-                               location=database.lineno(kwdargs.get('location')))
-
-    # Prevent loading the same file twice
-    if filename not in database.source_files:
-        identifier = len(database.source_files)
-        database.source_files.append(filename)
-        database.source_parent.append(kwdargs.get('location'))
-        pl = PrologFile(filename, identifier=identifier, factory=database.extra_info.get('factory'),
-                        parser=database.extra_info.get('parser'))
-        database.line_info.append(pl.line_info[0])
-        for clause in pl:
-            database += clause
-        # engine._process_directives(database)
-
+    database.consult(filename, location=kwdargs.get('location'))
     return True
 
 
@@ -1621,317 +1590,52 @@ def _builtin_findall(pattern, goal, result, **kwdargs):
     return _builtin_findall_base(pattern, goal, result, **kwdargs)
 
 
-# noinspection PyPep8Naming
-class problog_export(object):
-    database = None
-
-    @classmethod
-    def add_function(cls, name, in_args, out_args, func):
-        if problog_export.database is not None:
-            problog_export.database.add_extern(name, in_args + out_args, func)
-
-    # noinspection PyUnusedLocal
-    def __init__(self, *args, **kwdargs):
-        # TODO check if arguments are in order: input first, output last
-        self.input_arguments = [a[1:] for a in args if a[0] == '+']
-        self.output_arguments = [a[1:] for a in args if a[0] == '-']
-        self.functor = kwdargs.get('functor')
-
-    def _convert_input(self, a, t):
-        if t == 'str':
-            return str(a)
-        elif t == 'int':
-            return int(a)
-        elif t == 'float':
-            return float(a)
-        elif t == 'list':
-            return term2list(a)
-        elif t == 'term':
-            return a
-        elif t == 'obj':
-            return a.functor
-        else:
-            raise ValueError("Unknown type specifier '%s'!" % t)
-
-    def _type_to_callmode(self, t):
-        if t == 'str':
-            return 'a'
-        elif t == 'int':
-            return 'i'
-        elif t == 'float':
-            return 'f'
-        elif t == 'list':
-            return 'L'
-        elif t == 'term':
-            return '*'
-        elif t == 'obj':
-            return 'o'
-        else:
-            raise ValueError("Unknown type specifier '%s'!" % t)
-
-    def _extract_callmode(self):
-        callmode_in = ''
-        for t in self.input_arguments:
-            callmode_in += self._type_to_callmode(t)
-
-        # multiple call modes: index = binary encoding on whether the output is bound
-        # 0 -> all unbound
-        # 1 -> first output arg is bound
-        # 2 -> second output arg is bound
-        # 3 -> first and second are bound
-
-        n = len(self.output_arguments)
-        for i in range(0, 1 << n):
-            callmode = callmode_in
-            for j, t in enumerate(self.output_arguments):
-                if i & (1 << (n - j - 1)):
-                    callmode += self._type_to_callmode(t)
-                else:
-                    callmode += 'v'
-            yield callmode
-
-    def _convert_output(self, a, t):
-        if t == 'str':
-            return Term(a)
-        elif t == 'int':
-            return Constant(a)
-        elif t == 'float':
-            return Constant(a)
-        elif t == 'list':
-            return list2term(a)
-        elif t == 'term':
-            if not isinstance(a, Term):
-                return Term(a)
-                # raise ValueError("Expected term output, got '%s' instead." % type(a))
-            return a
-        elif t == 'obj':
-            if not isinstance(a, Object):
-                return Object(a)
-            else:
-                return a
-        else:
-            raise ValueError("Unknown type specifier '%s'!" % t)
-
-    def _convert_inputs(self, args):
-        return [self._convert_input(a, t) for a, t in zip(args, self.input_arguments)]
-
-    def _convert_outputs(self, args):
-        return [self._convert_output(a, t) for a, t in zip(args, self.output_arguments)]
-
-    def __call__(self, func, funcname=None):
-        if funcname is None:
-            if self.functor is None:
-                funcname = func.__name__
-            else:
-                funcname = self.functor
-
-        def _wrapped_function(*args, **kwdargs):
-            bound = check_mode(args, list(self._extract_callmode()), funcname, **kwdargs)
-            converted_args = self._convert_inputs(args)
-
-            argspec = inspect.getargspec(func)
-            if argspec.keywords is not None:
-                result = func(*converted_args, **kwdargs)
-            else:
-                result = func(*converted_args)
-            if len(self.output_arguments) == 1:
-                result = [result]
-
-            try:
-                transformed = []
-                for i, r in enumerate(result):
-                    r = self._convert_output(r, self.output_arguments[i])
-                    if bound & (1 << (len(self.output_arguments) - i - 1)):
-                        r = unify_value(r, args[len(self.input_arguments) + i], {})
-                    transformed.append(r)
-                result = args[:len(self.input_arguments)] + tuple(transformed)
-                return [result]
-            except UnifyError:
-                return []
-
-        problog_export.add_function(funcname, len(self.input_arguments),
-                                    len(self.output_arguments), _wrapped_function)
-        return func
-
-
-class problog_export_builtin(object):
-
-    def __init__(self, functor, arity, **kwargs):
-        self.functor = functor
-        self.arity = arity
-
-    def __call__(self, func):
-        problog_export.add_function(self.functor, self.arity, 0, func)
-        return func
-
-
-class problog_export_builtin(object):
-
-    def __init__(self, functor, arity, **kwargs):
-        self.functor = functor
-        self.arity = arity
-
-    def __call__(self, function):
-        problog_export.add_function(self.functor, self.arity, 0, function)
-        return function
-
-
-# noinspection PyPep8Naming
-class problog_export_raw(problog_export):
-
-    # noinspection PyUnusedLocal
-    def __init__(self, *args, **kwdargs):
-        problog_export.__init__(self, *args, **kwdargs)
-        self.input_arguments = [a[1:] for a in args]
-
-    def _convert_input(self, a, t):
-        if is_variable(a):
-            return None
-        else:
-            return problog_export._convert_input(self, a, t)
-
-    def _extract_callmode(self):
-        callmode_in = ''
-
-        # multiple call modes: index = binary encoding on whether the output is bound
-        # 0 -> all unbound
-        # 1 -> first output arg is bound
-        # 2 -> second output arg is bound
-        # 3 -> first and second are bound
-
-        n = len(self.input_arguments)
-        for i in range(0, 1 << n):
-            callmode = callmode_in
-            for j, t in enumerate(self.input_arguments):
-                if i & (1 << (n - j - 1)):
-                    callmode += self._type_to_callmode(t)
-                else:
-                    callmode += 'v'
-            yield callmode
-
-    def __call__(self, func, funcname=None):
-        if funcname is None:
-            if self.functor is None:
-                funcname = func.__name__
-            else:
-                funcname = self.functor
-
-        def _wrapped_function(*args, **kwdargs):
-            bound = check_mode(args, list(self._extract_callmode()), funcname, **kwdargs)
-            converted_args = self._convert_inputs(args)
-            results = []
-            for result in func(*converted_args, **kwdargs):
-                if len(result) == 2 and type(result[0]) == tuple:
-                    # Probabilistic
-                    result, p = result
-                    raise Exception('We don\'t support probabilistic yet!')
-                else:
-                    p = None
-
-                # result is always a list of tuples
-                try:
-                    transformed = []
-                    for i, r in enumerate(result):
-                        r = self._convert_output(r, self.input_arguments[i])
-                        if bound & (1 << i):
-                            r = unify_value(r, args[i], {})
-                        transformed.append(r)
-                    result = tuple(transformed)
-                    results.append(result)
-                except UnifyError:
-                    pass
-            return results
-
-        problog_export.add_function(funcname, len(self.input_arguments),
-                                    0, _wrapped_function)
-        return func
-
-
-# noinspection PyPep8Naming
-class problog_export_nondet(problog_export):
-    def __call__(self, func, funcname=None):
-        if funcname is None:
-            if self.functor is None:
-                funcname = func.__name__
-            else:
-                funcname = self.functor
-
-        def _wrapped_function(*args, **kwdargs):
-            bound = check_mode(args, list(self._extract_callmode()), funcname, **kwdargs)
-            converted_args = self._convert_inputs(args)
-            results = []
-            for result in func(*converted_args):
-                if len(self.output_arguments) == 1:
-                    result = [result]
-
-                try:
-                    transformed = []
-                    for i, r in enumerate(result):
-                        r = self._convert_output(r, self.output_arguments[i])
-                        if bound & (1 << (len(self.output_arguments) - i - 1)):
-                            r = unify_value(r, args[len(self.input_arguments) + i], {})
-                        transformed.append(r)
-                    result = args[:len(self.input_arguments)] + tuple(transformed)
-                    results.append(result)
-                except UnifyError:
-                    pass
-            return results
-
-        problog_export.add_function(funcname, len(self.input_arguments),
-                                    len(self.output_arguments), _wrapped_function)
-        return func
-
-
 # noinspection PyUnusedLocal
 def _builtin_module(name, predicates, **kwargs):
     return True
 
 
 # noinspection PyUnusedLocal
-def _builtin_use_module2(filename, predicates, **kwdargs):
-    return _builtin_use_module(filename, **kwdargs)
+def _builtin_use_module2(filename, predicates, database=None, location=None, **kwdargs):
+    database.use_module(filename, predicates, location=location)
+    return True
 
 
 def _builtin_use_module(filename, database=None, location=None, **kwdargs):
-    if filename.functor == 'library' and filename.arity == 1:
-        from . import library_paths
-        libname = _atom_to_filename(filename.args[0])
-        filename = None
-        for path in library_paths:
-            filename = os.path.join(path, libname)
-            if os.path.exists(filename + '.pl'):
-                filename += '.pl'
-                break
-            elif os.path.exists(filename + '.py'):
-                filename += '.py'
-                break
-    else:
-        root = database.source_root
-        if filename.location:
-            source_root = database.source_files[filename.location[0]]
-            if source_root:
-                root = os.path.dirname(source_root)
-
-        filename = os.path.join(root, _atom_to_filename(filename))
-
-    if filename is None:
-        raise ConsultError('Unknown library location', database.lineno(location))
-    elif filename is not None and filename[-3:] == '.py':
-        try:
-            load_external_module(database, filename)
-        except IOError as err:
-            raise ConsultError('Error while reading external library: %s' % str(err),
-                               database.lineno(location))
-        return True
-    else:
-        return _builtin_consult(Term(filename), database=database, location=location, **kwdargs)
+    database.use_module(filename, None, location=location)
+    return True
 
 
-def load_external_module(database, filename):
-    import imp
-    problog_export.database = database
-    with open(filename, 'r') as extfile:
-        imp.load_module('externals', extfile, filename, ('.py', 'U', 1))
+@builtin_boolean('_use_module', 3)
+def _builtin_use_module2_scope(scope, filename, predicates, database=None, location=None, **kwdargs):
+    scope = str(scope)
+    if scope == 'None':
+        scope = None
+    database.use_module(filename, predicates, my_scope=scope, location=location)
+    return True
+
+
+@builtin_boolean('_use_module', 2)
+def _use_module(scope, filename, database=None, location=None, **kwdargs):
+    scope = str(scope)
+    if scope == 'None':
+        scope = None
+    database.use_module(filename, None, my_scope=scope, location=location)
+    return True
+
+@builtin_boolean('_consult', 2)
+def _builtin_consult(scope, filename, database=None, engine=None, **kwdargs):
+    check_mode((filename,), ['a'], functor='consult', **kwdargs)
+    scope = str(scope)
+    if scope == 'None':
+        scope = None
+    database.consult(filename, location=kwdargs.get('location'), my_scope=scope)
+    return True
+
+@builtin_boolean('dbg_printdb', 0)
+def _dbg_printdb(database=None, **kwargs):
+    print(database, file=sys.stderr)
+    return True
 
 
 def _builtin_try_call(term, **kwdargs):
@@ -2176,3 +1880,14 @@ def seq(term, database=None, **kwargs):
 
     return [(Constant(s),)]
 
+@builtin_boolean('trace', 0)
+def trace(engine=None, **kwargs):
+    if engine.debugger:
+        engine.debugger.interactive = True
+    return True
+
+@builtin_boolean('notrace', 0)
+def notrace(engine=None, **kwargs):
+    if engine.debugger:
+        engine.debugger.interactive = False
+    return True
