@@ -158,7 +158,7 @@ class BaseFormula(ProbLogObject):
     # ==========                           NODE NAMES                            =========== #
     # ====================================================================================== #
 
-    def add_name(self, name, key, label=None):
+    def add_name(self, name, key, label=None, keep_name=False):
         """Add a name to the given node.
 
         :param name: name of the node
@@ -188,7 +188,7 @@ class BaseFormula(ProbLogObject):
     #     names = self.get_names()
     #     print (names)
 
-    def add_query(self, name, key):
+    def add_query(self, name, key, keep_name=False):
         """Add a query name.
 
         Same as ``add_name(name, key, self.LABEL_QUERY)``.
@@ -196,9 +196,9 @@ class BaseFormula(ProbLogObject):
         :param name: name of the query
         :param key: key of the query node
         """
-        self.add_name(name, key, self.LABEL_QUERY)
+        self.add_name(name, key, self.LABEL_QUERY, keep_name=keep_name)
 
-    def add_evidence(self, name, key, value):
+    def add_evidence(self, name, key, value, keep_name=False):
         """Add an evidence name.
 
         Same as ``add_name(name, key, self.LABEL_EVIDENCE_???)``.
@@ -208,11 +208,11 @@ class BaseFormula(ProbLogObject):
         :param value: value of the evidence (True, False or None)
         """
         if value is None:
-            self.add_name(name, key, self.LABEL_EVIDENCE_MAYBE)
+            self.add_name(name, key, self.LABEL_EVIDENCE_MAYBE, keep_name=keep_name)
         elif value:
-            self.add_name(name, key, self.LABEL_EVIDENCE_POS)
+            self.add_name(name, key, self.LABEL_EVIDENCE_POS, keep_name=keep_name)
         else:
-            self.add_name(name, key, self.LABEL_EVIDENCE_NEG)
+            self.add_name(name, key, self.LABEL_EVIDENCE_NEG, keep_name=keep_name)
 
     def clear_evidence(self):
         """Remove all evidence."""
@@ -440,12 +440,13 @@ class LogicFormula(BaseFormula):
     # ==========                         MANAGE LABELS                           =========== #
     # ====================================================================================== #
 
-    def add_name(self, name, key, label=None):
+    def add_name(self, name, key, label=None, keep_name=False):
         """Associates a name to the given node identifier.
 
             :param name: name of the node
             :param key: id of the node
             :param label: type of node (see LogicFormula.LABEL_*)
+            :param keep_name: keep name of node if it exists
         """
         if self._use_string_names:
             name = str(name)
@@ -457,11 +458,12 @@ class LogicFormula(BaseFormula):
                 lname = -name
             else:
                 lname = name
-            if ntype == 'atom':
-                node = type(node)(*(node[:-2] + (lname, node[-1])))
-            else:
-                node = type(node)(*(node[:-1] + (lname,)))
-            self._update(abs(key), node)
+            if not keep_name:
+                if ntype == 'atom':
+                    node = type(node)(*(node[:-2] + (lname, node[-1])))
+                else:
+                    node = type(node)(*(node[:-1] + (lname,)))
+                self._update(abs(key), node)
 
         BaseFormula.add_name(self, name, key, label)
 
@@ -590,7 +592,7 @@ class LogicFormula(BaseFormula):
                 node_id = self._add_constraint_me(group, node_id)
             return node_id
 
-    def add_and(self, components, key=None, name=None):
+    def add_and(self, components, key=None, name=None, compact=None):
         """Add a conjunction to the logic formula.
 
         :param components: a list of node identifiers that already exist in the logic formula.
@@ -598,9 +600,9 @@ class LogicFormula(BaseFormula):
         :param name: name of the node
         :returns: the key of the node in the formula (returns 0 for deterministic atoms)
         """
-        return self._add_compound('conj', components, self.FALSE, self.TRUE, key=key, name=name)
+        return self._add_compound('conj', components, self.FALSE, self.TRUE, key=key, name=name, compact=compact)
 
-    def add_or(self, components, key=None, readonly=True, name=None, placeholder=False):
+    def add_or(self, components, key=None, readonly=True, name=None, placeholder=False, compact=None):
         """Add a disjunction to the logic formula.
 
         :param components: a list of node identifiers that already exist in the logic formula.
@@ -622,7 +624,8 @@ class LogicFormula(BaseFormula):
         This may cause the data structure to contain superfluous nodes.
         """
         return self._add_compound('disj', components, self.TRUE, self.FALSE, key=key,
-                                  readonly=readonly and not placeholder, name=name, placeholder=placeholder)
+                                  readonly=readonly and not placeholder, name=name,
+                                  placeholder=placeholder, compact=compact)
 
     def add_disjunct(self, key, component):
         """Add a component to the node with the given key.
@@ -679,13 +682,13 @@ class LogicFormula(BaseFormula):
 
     # noinspection PyUnusedLocal
     def _add_compound(self, nodetype, content, t, f, key=None,
-                      readonly=True, update=None, name=None, placeholder=False):
+                      readonly=True, update=None, name=None, placeholder=False, compact=None):
         """Add a compound term (AND or OR)."""
         if not placeholder:
             assert content   # Content should not be empty
 
         name_clash = False
-        if self._auto_compact:
+        if compact or self._auto_compact and compact is None:
             # If there is a t node, (true for OR, false for AND)
             if t in content:
                 return t
@@ -812,8 +815,7 @@ class LogicFormula(BaseFormula):
             self.get_evidence_values()[key] = value
 
     def propagate(self, nodeids, current=None):
-        """Propagate the value of the given node
-          (true if node is positive, false if node is negative)
+        """Propagate the value of the given node (true if node is positive, false if node is negative)
         The propagation algorithm is not complete.
 
         :param nodeids: evidence nodes to set (> 0 means true, < 0 means false)
