@@ -1,3 +1,24 @@
+:- module(apply, [
+    include/3,
+    exclude/3,
+    partition/4,
+    partition/5,
+    maplist/2,
+    maplist/3,
+    maplist/4,
+    maplist/5,
+    convlist/3,
+    foldl/4,
+    foldl/5,
+    foldl/6,
+    foldl/7,
+    scanl/4,
+    scanl/5,
+    scanl/6,
+    scanl/7,
+    foldl_noinit/3
+]).
+
 
 include(P, [], []).
 include(P, [X|T], [X|S]) :-
@@ -23,7 +44,17 @@ partition(P, [X|R], S, [X|T]) :-
     \+ call(P, X),
     partition(P, R, S, T).
 
-% TODO implement partition/5
+partition(_, [], [], [], []).
+partition(P, [X|R], [X|S], T, U) :-
+    call(P, X, <),
+    partition(P, R, S, T, U).
+partition(P, [X|R], S, [X|T], U) :-
+    call(P, X, =),
+    partition(P, R, S, T, U).
+partition(P, [X|R], S, T, [X|U]) :-
+    call(P, X, >),
+    partition(P, R, S, T, U).
+
 
 maplist(_, []).
 maplist(P, [A|S]) :-
@@ -39,6 +70,14 @@ maplist(P, [A|S], [B|T]) :-
     call(P, A, B),
     maplist(P, S, T).
 
+convlist(_, [], []).
+convlist(P, [A|S], [B|T]) :-
+    call(P, A, B),
+    convlist(P, S, T).
+convlist(P, [A|S], T) :-
+    \+ call(P, A, _),
+    convlist(P, S, T).
+
 maplist(_, [], [], []).
 maplist(P, [A|S], [B|T], [C|R]) :-
     call(P, A, B, C),
@@ -103,3 +142,6 @@ scanl(_, [], [], [], [], _, []).
 scanl(P, [L1|LR1], [L2|LR2], [L3|LR3], [L4|LR4], L0, [R|RR]) :-
     call(P, L1, L2, L3, L4, L0, R),
     scanl(P, LR1, LR2, LR3, LR4, R, RR).
+
+
+
