@@ -257,6 +257,11 @@ def run_mpe_jsonp(model, callback):
     return run_problog_task('mpe', model[0], callback[0], options=['--full'])
 
 
+@handle_url(api_root + 'map')
+def run_map_jsonp(model, callback):
+    return run_problog_task('map', model[0], callback[0], options=[])
+
+
 @handle_url(api_root + 'sample')
 def run_sample_jsonp(model, callback):
     return run_problog_task('sample', model[0], callback[0])
@@ -376,6 +381,19 @@ def get_editor():
     return 200, 'text/html', data
 
 
+@handle_url(api_root + 'update')
+def update_problog():
+
+    workingdir = os.path.abspath(os.path.dirname(__file__))
+    try:
+        out = subprocess.check_output(['git', 'pull'], cwd=workingdir, stderr=subprocess.STDOUT).decode()
+        result = {'success': True, 'output': out}
+    except subprocess.CalledProcessError as err:
+        result = {'success': False, 'error': str(err.output)}
+
+    return 200, 'application/json', json.dumps(result)
+
+
 def make_local(html):
     import re
 
